---------------------------------------------------------------------------
--                                                                       --
--  Ast2Cfg is an ASIS application that generates a CFG out of the AST.  --
--                          http://cfg.w3x.org                           --
--                                                                       --
---------------------------------------------------------------------------
--                                                                       --
--     Copyright (c) 2007, 2008, 2009 Raul Fechete, Georg Kienesberger   --
--                                                                       --
---------------------------------------------------------------------------
--                                                                       --
-- This file is part of Ast2Cfg.                                         --
--                                                                       --
-- Ast2Cfg is free software; you can redistribute it and/or modify       --
-- it under the terms of the GNU General Public License as published by  --
-- the Free Software Foundation; either version 3 of the License, or     --
-- (at your option) any later version.                                   --
--                                                                       --
-- Ast2Cfg is distributed in the hope that it will be useful,            --
-- but WITHOUT ANY WARRANTY; without even the implied warranty of        --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         --
-- GNU General Public License for more details.                          --
--                                                                       --
-- You should have received a copy of the GNU General Public License     --
-- along with this program.  If not, see <http://www.gnu.org/licenses/>. --
--                                                                       --
---------------------------------------------------------------------------

with Ada.Exceptions;
with Ada.Characters.Handling;

with Ast2Cfg.Transformation;
with Ast2cfg.Output; use Ast2cfg.Output;
with Ast2Cfg.Flow_World;
with Ast2Cfg.Context_Processing;
with Ast2Cfg.Utils; use Ast2Cfg.Utils;
use Ast2cfg;

package body Ast2Cfg.Control is

   procedure Init (Context_Parameters: in Wide_String := "") is
   begin

      Transformation.Init;

      Asis.Implementation.Initialize(Initialization_Parameters);

      if Context_Parameters /= "" then
         Put_Line(Debug, "[Control] Context_Parameters: """ & W2s(Context_Parameters) & """");
         Asis.Ada_Environments.Associate
           (The_Context => My_Context,
            Name        => My_Context_Name,
            Parameters  => Context_Parameters);
      else
         Put_Line(Debug, "[Control] Context_Parameters: """ & W2s(My_Context_Parameters) & """");
         Asis.Ada_Environments.Associate
           (The_Context => My_Context,
            Name        => My_Context_Name,
            Parameters  => My_Context_Parameters);
      end if;

      Asis.Ada_Environments.Open(My_Context);

   exception
      when Ex : Asis.Exceptions.ASIS_Inappropriate_Context  |
        Asis.Exceptions.ASIS_Inappropriate_Container        |
        Asis.Exceptions.ASIS_Inappropriate_Compilation_Unit |
        Asis.Exceptions.ASIS_Inappropriate_Element          |
        Asis.Exceptions.ASIS_Inappropriate_Line             |
        Asis.Exceptions.ASIS_Inappropriate_Line_Number      |
        Asis.Exceptions.ASIS_Failed                         =>

         Put(Error, "[Ast2Cfg] ERROR: ASIS exception (");
         Put(Error, Ada.Exceptions.Exception_Name(Ex));
         Put(Error, ") is raised");
         New_Line(Error);

         Put(Error, "ASIS Error Status is ");
         Put(Error, Asis.Errors.Error_Kinds'Image(Asis.Implementation.Status));
         New_Line(Error);

         Put(Error, "ASIS Diagnosis is ");
         New_Line(Error);
         Put(Error, Ada.Characters.Handling.To_String(Asis.Implementation.Diagnosis));
         New_Line(Error);

         Asis.Implementation.Set_Status;

         raise;

      when Ex : others =>

         Put(Error, "[Ast2Cfg] ERROR: ");
         Put(Error, Ada.Exceptions.Exception_Name(Ex));
         Put(Error, " is raised (");
         Put(Error, Ada.Exceptions.Exception_Information(Ex));
         Put(Error, ")");
         New_Line(Error);

         raise;
   end Init;

   function Generate return Ast2Cfg.Flow_World.World_Object_Ptr is
   begin

      Context_Processing.Process_Context(The_Context => My_Context);
      World := Ast2Cfg.Transformation.Final;
      return World;

   exception
      when Ex : Asis.Exceptions.ASIS_Inappropriate_Context  |
        Asis.Exceptions.ASIS_Inappropriate_Container        |
        Asis.Exceptions.ASIS_Inappropriate_Compilation_Unit |
        Asis.Exceptions.ASIS_Inappropriate_Element          |
        Asis.Exceptions.ASIS_Inappropriate_Line             |
        Asis.Exceptions.ASIS_Inappropriate_Line_Number      |
        Asis.Exceptions.ASIS_Failed                         =>

         Put(Error, "[Ast2Cfg] ERROR: ASIS exception (");
         Put(Error, Ada.Exceptions.Exception_Name(Ex));
         Put(Error, ") is raised");
         New_Line(Error);

         Put(Error, "ASIS Error Status is ");
         Put(Error, Asis.Errors.Error_Kinds'Image(Asis.Implementation.Status));
         New_Line(Error);

         Put(Error, "ASIS Diagnosis is ");
         New_Line(Error);
         Put(Error, Ada.Characters.Handling.To_String(Asis.Implementation.Diagnosis));
         New_Line(Error);

         Asis.Implementation.Set_Status;

         raise;

      when Ex : others =>

         Put(Error, "[Ast2Cfg] ERROR: ");
         Put(Error, Ada.Exceptions.Exception_Name(Ex));
         Put(Error, " is raised (");
         Put(Error, Ada.Exceptions.Exception_Information(Ex));
         Put(Error, ")");
         New_Line(Error);

         raise;
   end Generate;

   procedure Final is
   begin

      Put_Line(Debug, "[Ast2Cfg] Finalizing Flow_World.");
      Ast2Cfg.Flow_World.Free(World);

      Put_Line(Debug, "[Ast2Cfg] Finalizing ASIS.");
      Asis.Ada_Environments.Close(My_Context);
      Asis.Ada_Environments.Dissociate(My_Context);
      Asis.Implementation.Finalize(Finalization_Parameters);

      Put_Line(Debug, "[Ast2Cfg] Ast2cfg finalized.");
   exception

      when Ex : Asis.Exceptions.ASIS_Inappropriate_Context  |
        Asis.Exceptions.ASIS_Inappropriate_Container        |
        Asis.Exceptions.ASIS_Inappropriate_Compilation_Unit |
        Asis.Exceptions.ASIS_Inappropriate_Element          |
        Asis.Exceptions.ASIS_Inappropriate_Line             |
        Asis.Exceptions.ASIS_Inappropriate_Line_Number      |
        Asis.Exceptions.ASIS_Failed                         =>

         Put(Error, "[Ast2Cfg] ERROR: ASIS exception (");
         Put(Error, Ada.Exceptions.Exception_Name(Ex));
         Put(Error, ") is raised");
         New_Line(Error);

         Put(Error, "ASIS Error Status is ");
         Put(Error, Asis.Errors.Error_Kinds'Image(Asis.Implementation.Status));
         New_Line(Error);

         Put(Error, "ASIS Diagnosis is ");
         New_Line(Error);
         Put(Error, Ada.Characters.Handling.To_String(Asis.Implementation.Diagnosis));
         New_Line(Error);

         Asis.Implementation.Set_Status;

         raise;

      when Ex : others =>

         Put(Error, "[Ast2Cfg] ERROR: ");
         Put(Error, Ada.Exceptions.Exception_Name(Ex));
         Put(Error, " is raised (");
         Put(Error, Ada.Exceptions.Exception_Information(Ex));
         Put(Error, ")");
         New_Line(Error);

         raise;
   end Final;

end Ast2Cfg.Control;
