---------------------------------------------------------------------------
--                                                                       --
--  Ast2Cfg is an ASIS application that generates a CFG out of the AST.  --
--                          http://cfg.w3x.org                           --
--                                                                       --
---------------------------------------------------------------------------
--                                                                       --
--     Copyright (c) 2007, 2008, 2009 Raul Fechete, Georg Kienesberger   --
--                                                                       --
---------------------------------------------------------------------------
--                                                                       --
-- This file is part of Ast2Cfg.                                         --
--                                                                       --
-- Ast2Cfg is free software; you can redistribute it and/or modify       --
-- it under the terms of the GNU General Public License as published by  --
-- the Free Software Foundation; either version 3 of the License, or     --
-- (at your option) any later version.                                   --
--                                                                       --
-- Ast2Cfg is distributed in the hope that it will be useful,            --
-- but WITHOUT ANY WARRANTY; without even the implied warranty of        --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         --
-- GNU General Public License for more details.                          --
--                                                                       --
-- You should have received a copy of the GNU General Public License     --
-- along with this program.  If not, see <http://www.gnu.org/licenses/>. --
--                                                                       --
---------------------------------------------------------------------------

with Ada.Strings.Fixed;
with Ada.Unchecked_Deallocation;
with Ada.Strings.Hash;

package body Ast2Cfg.Id is

   function New_Pkg_Id return Pointer is
      Ret: Pointer;
   begin
      if Pkgs = Long_Long_Natural'Last then
         Ada.Exceptions.Raise_Exception(Pkg_Overflow'Identity, "Maximum package count reached.");
      end if;

      Pkgs := Pkgs + 1;
      Cfgs := 0;
      Nodes := 0;

      Ret := new Object;

      return Ret;
   end New_Pkg_Id;

   function New_Cfg_Id return Pointer is
      Ret: Pointer;
   begin
      if Cfgs = Long_Long_Natural'Last then
         Ada.Exceptions.Raise_Exception(Cfg_Overflow'Identity, "Maximum CFG count for package number " &
                                        Ada.Strings.Fixed.Trim(Long_Long_Natural'Image(Pkgs), Ada.Strings.Both) & " reached.");
      end if;

      Cfgs := Cfgs + 1;
      Nodes := 0;

      Ret := new Object;

      return Ret;
   end New_Cfg_Id;

   function New_Node_Id return Pointer is
      Ret: Pointer;
   begin
      if Nodes = Long_Long_Natural'Last then
         Ada.Exceptions.Raise_Exception(Node_Overflow'Identity, "Maximum node count for CFG number " &
                                        Ada.Strings.Fixed.Trim(Long_Long_Natural'Image(Cfgs), Ada.Strings.Both) & " in package number " &
                                        Ada.Strings.Fixed.Trim(Long_Long_Natural'Image(Pkgs), Ada.Strings.Both) & " reached.");
      end if;

      Nodes := Nodes + 1;

      Ret := new Object;

      return Ret;
   end New_Node_Id;

   function  Get_Pkg_Num  (Id: in Pointer) return Long_Long_Natural is
   begin
      return Id.Pkg;
   end Get_Pkg_Num;

   function  Get_Cfg_Num  (Id: in Pointer) return Long_Long_Natural is
   begin
      return Id.Cfg;
   end Get_Cfg_Num;

   function  Get_Node_Num (Id: in Pointer) return Long_Long_Natural is
   begin
      return Id.Node;
   end Get_Node_Num;

   function "="(Left: in Pointer; Right: in Pointer) return Boolean is
   begin
      if (Get_Pkg_Num(Left) = Get_Pkg_Num(Right)) and
         (Get_Cfg_Num(Left) = Get_Cfg_Num(Right)) and
         (Get_Node_Num(Left) = Get_Node_Num(Right))   then
         return True;
      else
         return False;
      end if;
   end "=";

   function "<" (Left: in Pointer; Right: in Pointer) return Boolean is
   begin
      if Get_Pkg_Num(Left) < Get_Pkg_Num(Right) then
         return True;
      end if;

      if Get_Pkg_Num(Left) = Get_Pkg_Num(Right) then
         if Get_Cfg_Num(Left) < Get_Cfg_Num(Right) then
            return True;
         end if;

         if Get_Cfg_Num(Left) = Get_Cfg_Num(Right) then
            if Get_Node_Num(Left) < Get_Node_Num(Right) then
               return True;
            end if;
         end if;
      end if;

      return False;
   end "<";

   function ">" (Left: in Pointer; Right: in Pointer) return Boolean is
   begin
      if Get_Pkg_Num(Left) > Get_Pkg_Num(Right) then
         return True;
      end if;

      if Get_Pkg_Num(Left) = Get_Pkg_Num(Right) then
         if Get_Cfg_Num(Left) > Get_Cfg_Num(Right) then
            return True;
         end if;

         if Get_Cfg_Num(Left) = Get_Cfg_Num(Right) then
            if Get_Node_Num(Left) > Get_Node_Num(Right) then
               return True;
            end if;
         end if;
      end if;

      return False;
   end ">";

   function To_String (Id: in Pointer) return String is
   begin
      return Ada.Strings.Fixed.Trim(Long_Long_Natural'Image(Get_Pkg_Num(Id)), Ada.Strings.Both) & "." &
             Ada.Strings.Fixed.Trim(Long_Long_Natural'Image(Get_Cfg_Num(Id)), Ada.Strings.Both) & "." &
             Ada.Strings.Fixed.Trim(Long_Long_Natural'Image(Get_Node_Num(Id)), Ada.Strings.Both);
   end To_String;

   function To_CFG_Name (Id: in Pointer) return String is
   begin
      return "CFG[" & To_String(Id) & "]";
   end To_CFG_Name;

   function Hash (Id: in Pointer) return Ada.Containers.Hash_Type is
   begin
      return Ada.Strings.Hash(To_String(Id));
   end Hash;

   procedure Free (Id: in out Pointer) is
      procedure Destroy is new Ada.Unchecked_Deallocation(Object, Pointer);
   begin
      Destroy(Id);
   end Free;

end Ast2Cfg.Id;
