---------------------------------------------------------------------------
--                                                                       --
--  Ast2Cfg is an ASIS application that generates a CFG out of the AST.  --
--                          http://cfg.w3x.org                           --
--                                                                       --
---------------------------------------------------------------------------
--                                                                       --
--     Copyright (c) 2007, 2008, 2009 Raul Fechete, Georg Kienesberger   --
--                                                                       --
---------------------------------------------------------------------------
--                                                                       --
-- This file is part of Ast2Cfg.                                         --
--                                                                       --
-- Ast2Cfg is free software; you can redistribute it and/or modify       --
-- it under the terms of the GNU General Public License as published by  --
-- the Free Software Foundation; either version 3 of the License, or     --
-- (at your option) any later version.                                   --
--                                                                       --
-- Ast2Cfg is distributed in the hope that it will be useful,            --
-- but WITHOUT ANY WARRANTY; without even the implied warranty of        --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         --
-- GNU General Public License for more details.                          --
--                                                                       --
-- You should have received a copy of the GNU General Public License     --
-- along with this program.  If not, see <http://www.gnu.org/licenses/>. --
--                                                                       --
---------------------------------------------------------------------------

with Ada.Strings.Hash;

with Ast2Cfg.Flow_Types; use Ast2Cfg.Flow_Types;
with Ast2Cfg.Nodes;
with Ast2Cfg.Nodes; use type Ast2Cfg.Nodes.Node_Class_Ptr; use type Ast2Cfg.Nodes.Node_Mark;
with Ast2Cfg.Output; use Ast2Cfg.Output;
with Ast2Cfg.Id; use type Ast2Cfg.Id.Pointer;

package body Ast2Cfg.Loop_Refinement is

   procedure Refine (CFG: in out CFGs.CFG_Object'Class) is
      Root: Nodes.Node_Class_Ptr := Cfgs.Get_Head(CFG);
      BE_Cur: Vertex_Map.Cursor;
   begin

      -- initializations
      Node_Count := 0;
      if Dfs_Sorted /= null then
         Free(Dfs_Sorted);
      end if;
      Id_To_Dfs_Num.Clear(Id_Dfs);
      Vertex_Map.Clear(Backedges);
      Post_Trans_Processing.Node_Set.Clear(Dangling_Roots);

      Dfs_Sorted := new Node_Array(1 .. Cfgs.Get_Node_Sum(CFG));

      -- fill Dfs_Sorted and find backedges
      Dfs(Root, 0);

      -- find the loops with tarjan's algorithm and refine them
      Refine_Loops;

      -- debug output of results
      Put_Line(Max, "[Loop_Refinement] Node Mapping:");

      for I in Dfs_Sorted'Range loop
         Put_Line(Max, Id.To_String(Nodes.Get_Id(Nodes.Node_Object(Dfs_Sorted(I).all))) & " - " &
                  Ast2Cfg.Long_Long_Natural'Image(Id_To_Dfs_Num.Element(Id_Dfs,
                                                                        Nodes.Get_Id(Nodes.Node_Object(Dfs_Sorted(I).all)))));
      end loop;

      Put_Line(Max, "[Loop_Refinement] Backedges:");

      BE_Cur := Vertex_Map.First(Backedges);
      while BE_Cur /= Vertex_Map.No_Element loop
         Put_Line(Max, Ast2Cfg.Long_Long_Natural'Image(Vertex_Map.Element(BE_Cur)) &
                  " --> " & Ast2Cfg.Long_Long_Natural'Image(Vertex_Map.Key(BE_Cur)));
         Vertex_Map.Next(BE_Cur);
      end loop;

      -- remove dangling nodes; has to be the last statement!
      Post_Trans_Processing.Strip_Dangling_Nodes(CFG, Dangling_Roots);

   end Refine;

   -- "... if this doesn't work well then we can blame the vendor." John Barnes, Programming in Ada 2005
   function Hash (X: in Long_Long_Natural) return Ada.Containers.Hash_Type is
   begin
      return Ada.Strings.Hash(Ast2Cfg.Long_Long_Natural'Image(X));
   end Hash;

   procedure Dfs (Cur: in Nodes.Node_Class_Ptr; Last: in Long_Long_Natural) is
      Succs: Flow_Types.Flow_Class_Ptr_List.Object;
      Succ: Nodes.Node_Class_Ptr := null;
      Cur_Num: Ast2Cfg.Long_Long_Natural;
   begin

      -- check the node mark
      if Cur = null or else Nodes.Get_Mark(Cur.all) = 6 then
         return;
      elsif Nodes.Get_Mark(Cur.all) = 5 then
         -- we found a backedge
         Vertex_Map.Insert(Backedges, Id_To_Dfs_Num.Element(Id_Dfs, Nodes.Get_Id(Cur.all)), Last);
         if Last = 0 then
            Put_Line(Error, "[Post_Trans_Processing] Refine_Loops: ERROR: Added backedge starting at node number 0!");
         end if;
         return;
      end if;

      -- set the mark of the current node
      Nodes.Set_Mark(Cur.all, 5);

      -- increment the dfs number
      Node_Count := Node_Count + 1;

      -- add node pointer to sorted array
      Dfs_Sorted(Dfs_Sorted'First + Node_Count - 1) := Cur;

      -- map the node id to the dfs number
      Id_To_Dfs_Num.Insert(Id_Dfs, Nodes.Get_Id(Cur.all), Node_Count);

      -- recurse for all successors
      Cur_Num := Node_Count;
      Succs := Nodes.Get_Succs(Cur.all);
      Flow_Types.Flow_Class_Ptr_List.Reset(Succs);
      while Flow_Types.Flow_Class_Ptr_List.Has_Next(Succs) loop
         Flow_Types.Flow_Class_Ptr_List.Get_Next(Succs, Flow_Types.Flow_Class_Ptr(Succ));
         Dfs(Succ, Cur_Num);
      end loop;

      -- recursion for this node finished; mark it with 6
      Nodes.Set_Mark(Cur.all, 6);
   end Dfs;

   procedure Refine_Loops is

      package Uf is new Ast2Cfg.Union_Find (Long_Long_Natural, Dfs_Sorted'First, Dfs_Sorted'Last);

      Uf_Sets: Uf.Sets;
      Loop_Exits: Vertex_Set.Set;
      Edges_To_Remove: Edge_Set.Set;
      use type Edge_Set.Cursor;
      ETR_Cur: Edge_Set.Cursor;

      procedure Collapse (Loop_Body: Vertex_Set.Set; Loop_Header: Long_Long_Natural) is
         Cur: Vertex_Set.Cursor;
         Cur_Num: Long_Long_Natural;
         Cur_Node, First_After_Loop: Nodes.Node_Class_Ptr := null;
         Cur_Loop_Exits: Vertex_Set.Set;
         Succs: Flow_Class_Ptr_List.Object;
         Succ: Nodes.Node_Class_Ptr;
         After_Edge_Count: Long_Long_Natural := 0;
         Edge_To_Remove: Edge := Nil_Edge;
      begin

         -- we found a loop
         Put(Max, "[Loop_Refinement] Found a Loop: ");
         Put(Max, Long_Long_Natural'Image(Loop_Header) & " -");

         -- for every node in the loop body...
         Cur := Vertex_Set.First(Loop_Body);
         while Cur /= Vertex_Set.No_Element loop
            -- ...collapse it on the header; we may use link here (instead of union),
            -- since we only get canonical elements from Findloop
            Cur_Num := Vertex_Set.Element(Cur);
            Uf.Link(Uf_Sets, Cur_Num, Loop_Header);
            Put(Max, " " & Long_Long_Natural'Image(Cur_Num));

            -- collect loop exits for the current loop and those for outer loops
            Cur_Node := Dfs_Sorted(Cur_Num);
            if Cur_Node.all in Nodes.Exit_Jump_Node_Object'Class then
               if Nodes.Get_Target(Nodes.Exit_Jump_Node_Object(Cur_Node.all)) = "" then
                  Vertex_Set.Insert(Cur_Loop_Exits, Cur_Num);
               else
                  Vertex_Set.Insert(Loop_Exits, Cur_Num);
               end if;
            end if;

            Vertex_Set.Next(Cur);
         end loop;
         New_Line(Max);

         -- find the edge to the first statement after the loop...
         Succs := Nodes.Get_Succs(Dfs_Sorted(Loop_Header).all);
         Flow_Class_Ptr_List.Reset(Succs);
         while Flow_Class_Ptr_List.Has_Next(Succs) loop
            Flow_Class_Ptr_List.Get_Next(Succs, Flow_Class_Ptr(Succ));
            if not Vertex_Set.Contains(Loop_Body, Id_To_Dfs_Num.Element(Id_Dfs, Nodes.Get_Id(Succ.all))) then
               First_After_Loop := Succ;
               After_Edge_Count := After_Edge_Count + 1;
            end if;
         end loop;

         -- ...check if there is exactly one such edge
         if After_Edge_Count < 1 then
            Put_Line(Error, "[Loop_Refinement] ERROR: No after edge found!");
            return;
         elsif After_Edge_Count > 1 then
            Put_Line(Error, "[Loop_Refinement] ERROR: Too many after edges found!");
            return;
         end if;

         -- for a trivial exit node (exit; or exit LABEL;) there should be no edge
         -- pointing to the statement right after it
         declare
            Succs: Flow_Class_Ptr_List.Object;
            Succ, Cur_Ptr: Nodes.Node_Class_Ptr := null;
         begin
            if not Vertex_Set.Is_Empty(Cur_Loop_Exits) then
               Cur := Vertex_Set.First(Cur_Loop_Exits);
               while Cur /= Vertex_Set.No_Element loop
                  Cur_Ptr := Dfs_Sorted(Vertex_Set.Element(Cur));
                  if Cur_Ptr.all in Nodes.Trivial_Exit_Jump_Node_Object'Class then
                     Succs := Nodes.Get_Succs(Cur_Ptr.all);
                     if Flow_Class_Ptr_List.Count(Succs) /= 1 then
                        Put_Line(Error, "[Loop_Refinement] ERROR: Exit Node has too many successors!");
                        return;
                     end if;
                     Succ := Nodes.Node_Class_Ptr(Flow_Class_Ptr_List.Get_Head(Succs));
                     Edge_To_Remove.From := Cur_Ptr;
                     Edge_To_Remove.To := Succ;
                     Edges_To_Remove.Insert(Edge_To_Remove);
                     Edge_To_Remove := Nil_Edge;
                  end if;
                  Vertex_Set.Next(Cur);
               end loop;
            end if;
            if not Vertex_Set.Is_Empty(Loop_Exits) then
               Cur := Vertex_Set.First(Loop_Exits);
               while Cur /= Vertex_Set.No_Element loop
                  Cur_Ptr := Dfs_Sorted(Vertex_Set.Element(Cur));
                  if Cur_Ptr.all in Nodes.Trivial_Exit_Jump_Node_Object'Class then
                     Succs := Nodes.Get_Succs(Cur_Ptr.all);
                     if Flow_Class_Ptr_List.Count(Succs) /= 1 then
                        Put_Line(Error, "[Loop_Refinement] ERROR: Exit Node has too many successors!");
                        return;
                     end if;
                     Succ := Nodes.Node_Class_Ptr(Flow_Class_Ptr_List.Get_Head(Succs));
                     Edge_To_Remove.From := Cur_Ptr;
                     Edge_To_Remove.To := Succ;
                     Edges_To_Remove.Include(Edge_To_Remove);
                     Edge_To_Remove := Nil_Edge;
                  end if;
                  Vertex_Set.Next(Cur);
               end loop;
            end if;
         end;

         -- if we have just found loop exits for the current loop...
         if not Vertex_Set.Is_Empty(Cur_Loop_Exits) then
            Cur := Vertex_Set.First(Cur_Loop_Exits);
            while Cur /= Vertex_Set.No_Element loop
               -- ...link the loop exits to the first statement after the loop.
               Cur_Num := Vertex_Set.Element(Cur);
               Nodes.Add_Succ(Dfs_Sorted(Cur_Num).all, Flow_Class_Ptr(First_After_Loop));
               Nodes.Add_Pred(First_After_Loop.all, Flow_Class_Ptr(Dfs_Sorted(Cur_Num)));

               Vertex_Set.Next(Cur);
            end loop;
         end if;

         -- if the loop has a label, connect the exit nodes for this label
         if Nodes.Get_Name(Dfs_Sorted(Loop_Header).all) /= "" then
            Cur := Vertex_Set.First(Loop_Exits);
            while Cur /= Vertex_Set.No_Element loop

               Cur_Num := Vertex_Set.Element(Cur);
               if Nodes.Get_Target(Nodes.Exit_Jump_Node_Object(Dfs_Sorted(Cur_Num).all)) =
                  Nodes.Get_Name(Dfs_Sorted(Loop_Header).all) then
                  Nodes.Add_Succ(Dfs_Sorted(Cur_Num).all, Flow_Class_Ptr(First_After_Loop));
                  Nodes.Add_Pred(First_After_Loop.all, Flow_Class_Ptr(Dfs_Sorted(Cur_Num)));
                  Vertex_Set.Delete(Loop_Exits, Cur);
                  exit;
               end if;
               Vertex_Set.Next(Cur);
               if Cur = Vertex_Set.No_Element then
                  Put_Line(Debug, "[Loop_Refinement] No exit node for loop with label " &
                           Nodes.Get_Name(Dfs_Sorted(Loop_Header).all) & " found!");
               end if;
            end loop;
         end if;

         -- if this is a simple loop statement (loop ... end loop;) mark the
         -- edge from the loop header to the first statement after the loop as
         -- "to remove".
         if Dfs_Sorted(Loop_Header).all in Nodes.Infinite_Loop_Node_Object then
            Edge_To_Remove.From := Dfs_Sorted(Loop_Header);
            Edge_To_Remove.To := First_After_Loop;
            Edges_To_Remove.Insert(Edge_To_Remove);
            Edge_To_Remove := Nil_Edge;
         end if;

      end Collapse;

      procedure Find_Loop (Potential_Header: Long_Long_Natural) is
         Worklist, Loop_Body: Vertex_Set.Set;
         Canonical, Work, Pred_Num: Long_Long_Natural;
         Work_Cur: Vertex_Set.Cursor;
         Preds: Flow_Class_Ptr_List.Object;
         Pred: Nodes.Node_Class_Ptr;
      begin
         -- the potential header is a loop header only if it is the target of a backedge
         if Vertex_Map.Contains(Backedges, Potential_Header) then

            -- find the canonical element of the node where the backedge originates and add it to the worklist
            Uf.Find(Uf_Sets, Vertex_Map.Element(Backedges, Potential_Header), Canonical);
            Vertex_Set.Insert(Worklist, Canonical);

            -- pick an element from the worklist until it is empty
            Work_Cur := Vertex_Set.First(Worklist);
            while Work_Cur /= Vertex_Set.No_Element loop
               Work := Vertex_Set.Element(Work_Cur);
               Vertex_Set.Delete(Worklist, Work_Cur);

               -- add the current element to the loop body
               Vertex_Set.Insert(Loop_Body, Work);

               -- for each predecessor that is not origin of a backedge...
               Preds := Nodes.Get_Preds(Nodes.Node_Object(Dfs_Sorted(Work).all));
               Flow_Class_Ptr_List.Reset(Preds);
               while Flow_Class_Ptr_List.Has_Next(Preds) loop
                  Flow_Class_Ptr_List.Get_Next(Preds, Flow_Class_Ptr(Pred));
                  Pred_Num := Id_To_Dfs_Num.Element(Id_Dfs, Nodes.Get_Id(Pred.all));

                  if not Vertex_Map.Contains(Backedges, Work) or else
                     Vertex_Map.Element(Backedges, Work) /= Pred_Num then

                     -- ...find the canonical element, and add it to the worklist in case
                     -- it is not already in the loop body, and is different from the header
                     Uf.Find(Uf_Sets, Pred_Num, Canonical);
                     if not Vertex_Set.Contains(Loop_Body, Canonical) and
                        Canonical /= Potential_Header then
                        -- we do not have to check if it is already on the worklist, since
                        -- include does no harm in case it is already there
                        Vertex_Set.Include(Worklist, Canonical);
                     end if;
                  end if;
               end loop;
               Work_Cur := Vertex_Set.First(Worklist);
            end loop;
            -- in case we found a loop call Collapse
            if not Vertex_Set.Is_Empty(Loop_Body) then
               Collapse(Loop_Body, Potential_Header);
            end if;
         end if;
      end Find_Loop;

   begin
      Uf.Init_Sets(Uf_Sets);

      for I in reverse Dfs_Sorted'Range loop
         Find_Loop(Potential_Header => Id_To_Dfs_Num.Element(Id_Dfs, Nodes.Get_Id(Dfs_Sorted(I).all)));
      end loop;

      -- remove the edges we could not remove during the algo
      ETR_Cur := Edge_Set.First(Edges_To_Remove);
      while ETR_Cur /= Edge_Set.No_Element loop

         -- delete the succ
         Nodes.Rem_Succ(Edge_Set.Element(ETR_Cur).From.all, Flow_Class_Ptr(Edge_Set.Element(ETR_Cur).To));

         -- delete the pred
         Nodes.Rem_Pred(Edge_Set.Element(ETR_Cur).To.all, Flow_Class_Ptr(Edge_Set.Element(ETR_Cur).From));

         -- remember this node to remove possible dangling nodes later
         Post_Trans_Processing.Node_Set.Include(Dangling_Roots, Edge_Set.Element(ETR_Cur).To);

         Edge_Set.Next(ETR_Cur);
      end loop;

      -- check if all labeled loops were handled
      if not Vertex_Set.Is_Empty(Loop_Exits) then
         Put_Line(Error, "[Loop_Refinement] ERROR: Not all found labeled exit nodes were connected!");
      end if;

   end Refine_Loops;

end Ast2Cfg.Loop_Refinement;
