---------------------------------------------------------------------------
--                                                                       --
--  Ast2Cfg is an ASIS application that generates a CFG out of the AST.  --
--                          http://cfg.w3x.org                           --
--                                                                       --
---------------------------------------------------------------------------
--                                                                       --
--     Copyright (c) 2007, 2008, 2009 Raul Fechete, Georg Kienesberger   --
--                                                                       --
---------------------------------------------------------------------------
--                                                                       --
-- This file is part of Ast2Cfg.                                         --
--                                                                       --
-- Ast2Cfg is free software; you can redistribute it and/or modify       --
-- it under the terms of the GNU General Public License as published by  --
-- the Free Software Foundation; either version 3 of the License, or     --
-- (at your option) any later version.                                   --
--                                                                       --
-- Ast2Cfg is distributed in the hope that it will be useful,            --
-- but WITHOUT ANY WARRANTY; without even the implied warranty of        --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         --
-- GNU General Public License for more details.                          --
--                                                                       --
-- You should have received a copy of the GNU General Public License     --
-- along with this program.  If not, see <http://www.gnu.org/licenses/>. --
--                                                                       --
---------------------------------------------------------------------------

with Ada.Strings.Unbounded; use Ada.Strings.Unbounded;
with Ada.Unchecked_Deallocation;

with Asis;

with Ast2Cfg.Generic_Pointer_List;
with Ast2Cfg.Flow_Types; use Ast2Cfg.Flow_Types;
with Ast2Cfg.Id;
with Ast2Cfg.Utils; use Ast2Cfg.Utils;

package Ast2Cfg.Nodes is

   type Node_Object     is new Flow_Object with private;
   type Node_Class_Ptr  is access all Node_Object'Class;

   -- Multi-purpose-Mark with a Value out of 0..7, -1 for reset
   type Node_Mark       is new Integer range -1..8;

   not overriding
   procedure Free (Node: in out Node_Class_Ptr);

   package Node_Class_Ptr_List is new Generic_Pointer_List (Item_Pointer => Node_Class_Ptr,
                                                            Free_Procedure => Free);

   not overriding -- for overriding Init see private part
   procedure Init (Node: in out Node_Object; Graph: in Flow_Class_Ptr := null; Element: in Asis.Element);

   not overriding
   procedure Free_Special (Node: in out Node_Class_Ptr);

   not overriding
   procedure Free_Tree_Recursive (Node: in out Node_Class_Ptr);

   not overriding
   procedure Free_Recursive (Node: in out Node_Class_Ptr);

   not overriding
   procedure Add_Ident  (Node: in out Node_Object; Identifier: in out UStr_List.Object);

   not overriding
   procedure Add_Idents (Node: in out Node_Object; Identifiers: in out UStr_List.Object);

   not overriding
   function  Get_Idents (Node: in Node_Object) return UStr_List_List.Object;

   not overriding
   procedure Set_Info (Node: in out Node_Object; Info: Unbounded_String);

   not overriding
   procedure Add_Info (Node: in out Node_Object; Info: Unbounded_String);

   not overriding
   function  Get_Info (Node: in Node_Object) return String;

   not overriding
   procedure Append_Info (Node: in out Node_Object; Info: Unbounded_String);

   -- Multi-purpose-Mark with a Value out of 0..8, -1 for reset
   not overriding
   procedure Set_Mark (Node: in out Node_Object; Mark: in Node_Mark);

   not overriding
   function  Get_Mark (Node: in Node_Object) return Node_Mark;

   -- resets the marks of all reachable nodes to 0 temporarily using -1
   not overriding
   procedure Reset_Mark (Node: in out Node_Class_Ptr);

   -- RTree
   not overriding
   procedure Set_RHS_Tree (Node: in out Node_Object; Tree: in Node_Class_Ptr);

   not overriding
   function  Get_RHS_Tree (Node: in Node_Object) return Node_Class_Ptr;

   -- Call Node
   type Call_Node_Object       is new Node_Object with private;

   overriding
   procedure Init(Node: in out Call_Node_Object; Graph: in Flow_Class_Ptr := null; Element: in Asis.Element);

   procedure Set_Call_Target(Node: in out Call_Node_Object; Element: in Asis.Element);
   function Get_Call_Target(Node: in Call_Node_Object) return Asis.Element;

   procedure Set_Call_Destination(Node: in out Call_Node_Object; Flow: Flow_Class_Ptr);
   function Get_Call_Destination(Node: in Call_Node_Object) return Flow_Class_Ptr;

   -- CFG to be called (e.g. Ada + Text_IO + Put)
   not overriding
   procedure Add_Call_Item  (Node: in out Call_Node_Object; Name: in Unbounded_String);

   not overriding
   function Get_Dest_Name (Node: in Call_Node_Object) return String;

   -- Accept Node
   type Accept_Node_Object     is new Call_Node_Object with private;

   -- Entry Node
   type Entry_Call_Node_Object is new Call_Node_Object with private;

   -- Abort Node
   type Abort_Node_Object      is new Call_Node_Object with private;

   -- Jump Node
   type Jump_Node_Object       is abstract new Node_Object with private;

   overriding
   procedure Init (Node: in out Jump_Node_Object; Graph: in Flow_Class_Ptr := null; Element: in Asis.Element);

   not overriding
   procedure Set_Target (Node: in out Jump_Node_Object; Target: in Unbounded_String);

   not overriding
   function  Get_Target (Node: in Jump_Node_Object) return String;

   type Goto_Jump_Node_Object  is new Jump_Node_Object with private;
   type Exit_Jump_Node_Object  is abstract new Jump_Node_Object with private;
   type Trivial_Exit_Jump_Node_Object  is new Exit_Jump_Node_Object with private;
   type Complex_Exit_Jump_Node_Object  is new Exit_Jump_Node_Object with private;

   -- If/Case Node
   type Branch_Node_Object     is new Node_Object with private;

   -- Return Node
   type Return_Node_Object     is new Node_Object with private;

   -- Terminate Node
   type Terminate_Node_Object  is new Node_Object with private;

   -- Loop Node
   type Loop_Node_Object          is abstract new Node_Object with private;
   type Finite_Loop_Node_Object   is new Loop_Node_Object with private;
   type Infinite_Loop_Node_Object is new Loop_Node_Object with private;

   -- Parameter Node
   type Param_Node_Object       is new Node_Object with private;
   type Param_Alloc_Node_Object is new Node_Object with private;
   type Param_Call_Node_Object  is new Call_Node_Object with private;

   not overriding -- for overriding Init see private part
   procedure Init (Node: in out Param_Node_Object; Graph: in Flow_Class_Ptr := null);

   not overriding -- for overriding Init see private part
   procedure Init (Node: in out Param_Alloc_Node_Object; Graph: in Flow_Class_Ptr := null);

   overriding -- for overriding Init see private part
   procedure Init (Node: in out Param_Call_Node_Object; Graph: in Flow_Class_Ptr := null; Element: in Asis.Element);

   not overriding
   procedure Add_Var_Item (Node: in out Param_Node_Object; Name: in Unbounded_String);
   not overriding
   function Get_Var_Name  (Node: in Param_Node_Object) return String;
   not overriding
   procedure Clear_Var    (Node: in out Param_Node_Object);

   not overriding
   procedure Add_Type_Item (Node: in out Param_Alloc_Node_Object; Name: in Unbounded_String);
   not overriding
   function Get_Type_Name  (Node: in Param_Alloc_Node_Object) return String;
   not overriding
   procedure Clear_Type    (Node: in out Param_Alloc_Node_Object);

   -- Assignment Node
   type Assign_Node_Object     is new Node_Object with private;

   overriding
   procedure Init (Node: in out Assign_Node_Object; Graph: in Flow_Class_Ptr := null; Element: in Asis.Element);

   not overriding
   procedure Set_LHS_Tree (Node: in out Assign_Node_Object; Tree: in Node_Class_Ptr);

   not overriding
   function  Get_LHS_Tree (Node: in Assign_Node_Object) return Node_Class_Ptr;

private

   type Node_Object is new Flow_Object with
      record
         Graph:   Flow_Class_Ptr;        -- Parent CFG
         Info:    Unbounded_String;      -- Program Code
         Mark:    Node_Mark;             -- Multi-Purpose Mark: Integer Range 0..7, -1 for reset
         RTree:   Node_Class_Ptr;        -- Function/Parameter Tree (RHS)
         Idents:  UStr_List_List.Object; -- List of the Found Identifiers
      end record;

   overriding
   procedure Init (Node: in out Node_Object;
                   Oid:  in Id.Pointer;
                   Name: in Unbounded_String;
                   Pred: in Flow_Class_Ptr := null;
                   Element: in Asis.Element := Asis.Nil_Element);

   -- Call Node

   type Call_Node_Object is new Node_Object with
      record
         Dest_Name: UStr_List.Object;
         Dest_Ptr:  Flow_Class_Ptr;
         Dest_Elem: Asis.Element;
      end record;

   -- Abort Node

   type Abort_Node_Object is new Call_Node_Object with
      record
         null;
      end record;

   -- Accept Node

   type Accept_Node_Object is new Call_Node_Object with
      record
         null;
      end record;

   -- Entry Call Node

   type Entry_Call_Node_Object is new Call_Node_Object with
      record
         null;
      end record;

   -- Jump Node

   type Jump_Node_Object is new Node_Object with
      record
         Target: Unbounded_String;
      end record;

   type Goto_Jump_Node_Object is new Jump_Node_Object with
     null record;

   type Exit_Jump_Node_Object is abstract new Jump_Node_Object with
     null record;

   type Trivial_Exit_Jump_Node_Object is new Exit_Jump_Node_Object with
     null record;

   type Complex_Exit_Jump_Node_Object is new Exit_Jump_Node_Object with
     null record;

   -- Branch Node

   type Branch_Node_Object is new Node_Object with
      record
         null;
      end record;

   -- Return Node

   type Return_Node_Object is new Node_Object with
      record
         null;
      end record;

   -- Terminate Call Node

   type Terminate_Node_Object is new Node_Object with
      record
         null;
      end record;

   -- Loop Node

   type Loop_Node_Object is new Node_Object with
     null record;

   type Finite_Loop_Node_Object is new Loop_Node_Object with
     null record;

   type Infinite_Loop_Node_Object is new Loop_Node_Object with
     null record;

   -- Parameter Node

   type Param_Node_Object is new Node_Object with
      record
         Var_Name: UStr_List.Object;
      end record;

   overriding
   procedure Init (Node: in out Param_Node_Object; Graph: in Flow_Class_Ptr := null; Element: in Asis.Element);

   type Param_Alloc_Node_Object is new Node_Object with
      record
         Type_Name: UStr_List.Object;
      end record;

   overriding
   procedure Init (Node: in out Param_Alloc_Node_Object; Graph: in Flow_Class_Ptr := null; Element: in Asis.Element);

   type Param_Call_Node_Object is new Call_Node_Object with
     null record;

   -- Assignment Node

   type Assign_Node_Object is new Node_Object with
      record
         LTree: Node_Class_Ptr;
      end record;

end Ast2Cfg.Nodes;
